/*=====================================================================*/
/*    serrano/prgm/project/bigloo/fthread/src/BglThread.java           */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Wed Jan 30 10:40:23 2002                          */
/*    Last change :  Wed Jul 10 14:06:57 2002 (serrano)                */
/*    Copyright   :  2002 Manuel Serrano                               */
/*    -------------------------------------------------------------    */
/*    Bigloo Class for threads. Nothing fancy here. Bigloo threads are */
/*    regular threads augmented with two slots:                        */
/*      - a slot used to store a back pointer to the Bigloo objects    */
/*        representing the user thread.                                */
/*      - a slot pointing to the closure implementing the body of the  */
/*        thread.                                                      */
/*=====================================================================*/

/*---------------------------------------------------------------------*/
/*    The package                                                      */
/*---------------------------------------------------------------------*/
package bigloo.fairthread;
import java.lang.*;
import bigloo.*;

/*---------------------------------------------------------------------*/
/*    BglThread                                                        */
/*---------------------------------------------------------------------*/
public class BglThread extends bigloo.bglthread {
   // Bigloo connection
   private Object bgl_obj;
   private procedure bgl_thunk;

   // Public lock
   public Object lock;
   
   // parent thread (i.e. the Java thread that has entered this Bgl thread)
   Thread parent;
   
   // wait/notify synchronization
   private boolean hascpup = false;
   private boolean started = false;

   // public constructor
   public BglThread() {
      super();
      lock = new Object();
   }
   
   // public constructor
   public BglThread( procedure thunk ) {
      super();
      lock = new Object();
      bgl_thunk = thunk;
   }

   // This method is invoked by Bigloo when the Bigloo object
   // representing the thread has been allocated.
   public void setup( Object o ) {
      bgl_obj = o;
   }

   // Return back the pointer to the Bigloo object
   public Object getBglObj() {
      return bgl_obj;
   }

   // Return back the body of the Java thread
   public Object getThunk() {
      return bgl_thunk;
   }

   // Run the thread
   public void run() {
      try {
	 hascpup = true;

	 synchronized( this ) {
	    started = true;
	    wait();
	    bgl_thunk.funcall0();
	 }
      } catch( Throwable e ) {
	 try {
	    foreign.internalerror( e );
	 } catch( Throwable _ ) {
	    System.exit( 1 );
	 }
      }
   }

   // await a notification
   public void await( BglThread s ) {
      synchronized( s.lock ) {
	 try{
	    s.lock.wait();
	 } catch( InterruptedException e ) { 
	    System.err.println( "interrupted wait exception: " + e ); 
	 }
      }
   }
   
   // private method to get the CPU
   private void getCpu() {
      hascpup = true;
      notify();
   }
   
   // private method to loose the CPU
   private void looseCpu() {
      while( hascpup == false ) {
	 try{
	    wait();
	 } catch( InterruptedException e ) { 
	    System.err.println( "interrupted wait exception: " + e ); 
	 }
      }
   }
   
   // Switch to another Bglthread
   public void switchTo( BglThread next ) {
      hascpup = false;
      
      while( next.started == false ) Thread.yield();

      synchronized( this ) {
	 synchronized( next ) {
	    next.getCpu();
	 }
	 looseCpu();
      }
   }
   
   // Enter a Bglthread from a regular thread
   static public void switchIn( BglThread next ) {
      next.parent = currentThread();
      
      while( next.started == false ) Thread.yield();

      synchronized( next.parent ) {
	 synchronized( next ) {
	    next.getCpu();
	 }
	 try{
	    next.parent.wait();
	 } catch( InterruptedException e ) { 
	    System.err.println( "interrupted wait exception: " + e ); 
	 }
      }
   }
   
   // Leave a Bglthread for a regular thread
   public void switchOut() {
      synchronized( parent ) {
	 parent.notify();
      }
      synchronized( this ) {
	 try{
	    wait();
	 } catch( InterruptedException e ) { 
	    System.err.println( "interrupted wait exception: " + e );
	 }
      }
   }
}
