;*=====================================================================*/
;*    serrano/prgm/project/bigloo/fthread/src/_thread.scm              */
;*    -------------------------------------------------------------    */
;*    Author      :  Frederic Boussinot                                */
;*    Creation    :  Mon Jan  7 13:50:54 2002                          */
;*    Last change :  Mon Mar  4 15:01:23 2002 (serrano)                */
;*    Copyright   :  2002 F. Boussinot                                 */
;*    -------------------------------------------------------------    */
;*    Fair threads                                                     */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __ft_%thread
   
   (import __ft_types
	   __ft_%types
	   __ft_signal)
   
   (java   (class %jobject "java.lang.Object")
	   
	   (class %native-thread::%jobject
	      (method start::void (::%native-thread)
		      "start")
	      "java.lang.Thread")
	   
	   (class %bgl-asynchronous-thunk::%native-thread
	      (constructor new (::procedure))
	      (method start::void (::%bgl-asynchronous-thunk)
		      "start")
	      (method static funcall::obj (::procedure ::obj)
		      "funcall")
	      (method static await!::void (::obj)
		      "await")
	      "bigloo.fairthread.BglAsyncThunk")
	   
	   (class %bgl-thread::%native-thread
	      (constructor new (::procedure))
	      (method setup!::void (::%bgl-thread ::obj) "setup")
	      (method obj::obj (::%bgl-thread) "getBglObj")
	      (method switch::void (::%bgl-thread ::%bgl-thread) "switchTo")
	      (method static switch-in::void (::%bgl-thread) "switchIn")
	      (method switch-out::void (::%bgl-thread) "switchOut")
	      "bigloo.fairthread.BglThread"))

   (extern (include "bglthread.h")
	   (type %bgl-thread (opaque) "bglthread_t")
	   (macro %native-thread-start::void (::%bgl-thread)
		  "bglthread_start")
	   (macro %bgl-thread-new::%bgl-thread (::procedure)
		  "bglthread_new")
	   (macro %bgl-thread-setup!::void (::%bgl-thread ::obj)
		  "bglthread_setup")
	   (macro %bgl-thread-switch::void (::%bgl-thread ::%bgl-thread)
		  "bglthread_switch_to")
	   (macro %bgl-thread-switch-in::void (::%bgl-thread)
		  "bglthread_switch_in")
	   (macro %bgl-thread-switch-out::void (::%bgl-thread)
		  "bglthread_switch_out")
	   (macro %bgl-asynchronous-thunk-new::%bgl-thread (::procedure)
		  "bglthread_async_new")
	   (macro %bgl-asynchronous-thunk-start::void (::%bgl-thread)
		  "bglthread_async_start")
	   (macro %bgl-asynchronous-thunk-funcall::obj (::procedure ::obj)
		  "bglthread_async_funcall")
	   (macro %bgl-asynchronous-thunk-await!::void (::obj)
		  "bglthread_async_await"))
   
   (export (thread-setup! ::thread)
	   
	   (thread-switch ::thread ::thread)
	   (thread-switch-in ::thread)
	   (thread-switch-out ::thread)
	   
	   (thread-attached? ::thread)
	   (thread-dead? ::thread)
	   (thread-cooperate ::thread)
	   
	   (thread-set-runnable! ::thread)
	   (thread-set-eoi! ::thread)
	   (thread-unset-eoi! ::thread)
	   (thread-set-await! ::thread)
	   (thread-set-yield! ::thread)
	   (thread-set-dead! ::thread)

	   (thread-add-mutex! ::thread ::mutex)
	   (thread-del-mutex! ::thread ::mutex)

	   (sort-threads::pair-nil ::pair-nil)))

;*---------------------------------------------------------------------*/
;*    *thread-counter* ...                                             */
;*---------------------------------------------------------------------*/
(define *thread-counter* 0)

;*---------------------------------------------------------------------*/
;*    thread-setup! ...                                                */
;*---------------------------------------------------------------------*/
(define (thread-setup! thread)
   (set! *thread-counter* (+fx 1 *thread-counter*))
   (with-access::thread thread (%builtin %ident)
      (%bgl-thread-setup! %builtin thread)
      (set! %ident *thread-counter*)
      thread))

;*---------------------------------------------------------------------*/
;*    thread-switch ...                                                */
;*---------------------------------------------------------------------*/
(define (thread-switch from to)
   (%bgl-thread-switch (thread-%builtin from) (thread-%builtin to))
   to)

;*---------------------------------------------------------------------*/
;*    thread-switch-in ...                                             */
;*---------------------------------------------------------------------*/
(define (thread-switch-in to)
   (%bgl-thread-switch-in (thread-%builtin to))
   to)

;*---------------------------------------------------------------------*/
;*    thread-switch-out ...                                            */
;*---------------------------------------------------------------------*/
(define (thread-switch-out from)
   (%bgl-thread-switch-out (thread-%builtin from))
   from)

;*---------------------------------------------------------------------*/
;*    thread-attached? ...                                             */
;*---------------------------------------------------------------------*/
(define (thread-attached? t)
   (thread-scheduler t))

;*---------------------------------------------------------------------*/
;*    thread-dead? ...                                                 */
;*---------------------------------------------------------------------*/
(define (thread-dead? t)
   (with-access::thread t (scheduler %state)
      (eq? %state 'dead)))

;*---------------------------------------------------------------------*/
;*    thread-cooperate ...                                             */
;*---------------------------------------------------------------------*/
(define (thread-cooperate t)
   (with-access::thread t (%state scheduler %terminate)
      (thread-switch t scheduler)
      (if (eq? %state 'dying)
	  (%terminate t))))

;*---------------------------------------------------------------------*/
;*    thread-set-runnable! ...                                         */
;*---------------------------------------------------------------------*/
(define (thread-set-runnable! t)
   (with-access::thread t (scheduler %state)
      ;; we have to take care to avoid resurection
      (if (not (memq %state '(dying dead)))
	  (begin
	     (set! %state 'runnable)
	     (with-access::%scheduler scheduler (threads-runnable)
		(set! threads-runnable (cons t threads-runnable)))))))

;*---------------------------------------------------------------------*/
;*    thread-set-eoi! ...                                              */
;*---------------------------------------------------------------------*/
(define (thread-set-eoi! t)
   (with-access::thread t (scheduler %state)
      (set! %state 'eoi)
      (with-access::%scheduler scheduler (threads-eoi)
	 (set! threads-eoi (cons t threads-eoi)))))

;*---------------------------------------------------------------------*/
;*    thread-unset-eoi! ...                                            */
;*---------------------------------------------------------------------*/
(define (thread-unset-eoi! t)
   (with-access::thread t (scheduler %state)
      (set! %state 'runnable)
      (with-access::%scheduler scheduler (threads-eoi)
	 (set! threads-eoi (remq! t threads-eoi)))))

;*---------------------------------------------------------------------*/
;*    thread-set-await! ...                                            */
;*---------------------------------------------------------------------*/
(define (thread-set-await! t)
   (with-access::thread t (%state)
      (set! %state 'await)))

;*---------------------------------------------------------------------*/
;*    thread-set-yield! ...                                            */
;*---------------------------------------------------------------------*/
(define (thread-set-yield! t)
   (with-access::thread t (scheduler %state)
      (set! %state 'yield)
      (with-access::%scheduler scheduler (threads-yield)
	 (set! threads-yield (cons t threads-yield)))))

;*---------------------------------------------------------------------*/
;*    thread-set-dead! ...                                             */
;*---------------------------------------------------------------------*/
(define (thread-set-dead! t)
   (with-access::thread t (scheduler %state)
      (set! %state 'dead)
      (with-access::%scheduler scheduler (threads-runnable threads-eoi deadn)
	 (set! deadn (+fx 1 deadn))
	 (set! threads-runnable (remq! t threads-runnable))
	 (set! threads-eoi (remq! t threads-eoi)))))

;*---------------------------------------------------------------------*/
;*    thread-add-mutex! ...                                            */
;*---------------------------------------------------------------------*/
(define (thread-add-mutex! t m)
   (with-access::thread t (%mutexes)
      (set! %mutexes (cons t %mutexes))))

;*---------------------------------------------------------------------*/
;*    thread-del-mutex! ...                                            */
;*---------------------------------------------------------------------*/
(define (thread-del-mutex! t m)
   (with-access::thread t (%mutexes)
      (set! %mutexes (remq! t %mutexes))))

;*---------------------------------------------------------------------*/
;*    sort-threads ...                                                 */
;*---------------------------------------------------------------------*/
(define (sort-threads threads)
   (if *thread-strict-order*
       (sort threads
	     (lambda (t1 t2)
		(<fx (thread-%ident t1)
		     (thread-%ident t2))))
       threads))
