/*=====================================================================*/
/*    serrano/prgm/project/bigloo/fthread/src/bglasyncthunk.c          */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Mon Feb 25 18:13:15 2002                          */
/*    Last change :  Wed Jul 24 11:29:28 2002 (serrano)                */
/*    Copyright   :  2002 Manuel Serrano                               */
/*    -------------------------------------------------------------    */
/*    The implementation of asynchronous thunk (non blocking IO).      */
/*=====================================================================*/
#include <pthread.h>
#include <stdlib.h>
#include <gc.h>

#define GC_PRIVATE_H
#include <bglthread.h>

/*---------------------------------------------------------------------*/
/*    Global lock                                                      */
/*---------------------------------------------------------------------*/
pthread_mutex_t async_lock;
pthread_cond_t async_cv;
static obj_t async_token = 0L;
extern char *strerror();

extern pthread_key_t bglkey;

/*---------------------------------------------------------------------*/
/*    bglthread_t                                                      */
/*    bglthread_async_new ...                                          */
/*---------------------------------------------------------------------*/
bglthread_t
bglthread_async_new( obj_t proc ) {
   bglthread_t thread = bglthread_new( proc );

   return thread;
}

/*---------------------------------------------------------------------*/
/*    static void *                                                    */
/*    bglthread_async_run ...                                          */
/*---------------------------------------------------------------------*/
static void *
bglthread_async_run( void *arg ) {
   bglthread_t thread = (bglthread_t)arg;

   thread->started = 1;
   
   pthread_setspecific( bglkey, arg );

   PROCEDURE_ENTRY( thread->thunk )( thread->thunk, BEOA );

   return BUNSPEC;
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bglthread_async_start ...                                        */
/*---------------------------------------------------------------------*/
void
bglthread_async_start( bglthread_t thread ) {
   static int num = 0;

   if( pthread_create( &(thread->pthread), 0L, bglthread_async_run, thread ) )
      FAILURE( string_to_bstring( "thread-start!" ),
	       string_to_bstring( "Cannot start thread" ),
	       string_to_bstring( strerror( errno ) ) );
   if( pthread_detach( thread->pthread ) )
      FAILURE( string_to_bstring( "thread-start!" ),
	       string_to_bstring( "Cannot start thread" ),
	       string_to_bstring( strerror( errno ) ) );
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    bglthread_async_funcall ...                                      */
/*---------------------------------------------------------------------*/
obj_t
bglthread_async_funcall( obj_t thunk, obj_t token ) {
   obj_t res;
   
   pthread_mutex_lock( &async_lock );
   
   async_token = 0L;
   res = PROCEDURE_ENTRY( thunk )( thunk, BEOA );
   async_token = token;
   
   pthread_cond_signal( &async_cv );
   pthread_mutex_unlock( &async_lock );
   
   return res;
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bglthread_async_await ...                                        */
/*---------------------------------------------------------------------*/
void
bglthread_async_await( obj_t t ) {
   while( async_token != t ) {
      pthread_cond_wait( &async_cv, &async_lock );
   }
}
				
