/*=====================================================================*/
/*    serrano/prgm/project/bigloo/fthread/src/bglrts.c                 */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Tue Feb 26 15:15:54 2002                          */
/*    Last change :  Wed Aug  7 11:05:46 2002 (serrano)                */
/*    Copyright   :  2002 Manuel Serrano                               */
/*    -------------------------------------------------------------    */
/*    This file implements the re-definition of standard Bigloo        */
/*    runtime facilities for thread safety. Mainly, it contains        */
/*    new definitions of all the Bigloo rts features that deal         */
/*    with the stack (such as errors, call/cc, and so on.)             */
/*=====================================================================*/
#include <pthread.h>
#include <stdlib.h>
#include <gc.h>

#define GC_PRIVATE_H
#include <bglthread.h>

/*---------------------------------------------------------------------*/
/*    Private variables                                                */
/*---------------------------------------------------------------------*/
obj_t (*bgl_orig_get_exitd_top)();
obj_t (*bgl_orig_set_exitd_top)( obj_t );

char *(*bgl_orig_get_stack_bottom)();
void (*bgl_orig_set_stack_bottom)( char * );

obj_t (*bgl_orig_get_current_output_port)();
obj_t (*bgl_orig_set_current_output_port)( obj_t );
obj_t (*bgl_orig_get_current_error_port)();
obj_t (*bgl_orig_set_current_error_port)( obj_t );
obj_t (*bgl_orig_get_current_input_port)();
obj_t (*bgl_orig_set_current_input_port)( obj_t );

int (*bgl_orig_get_mvalues_number)();
int (*bgl_orig_set_mvalues_number)( int );
obj_t (*bgl_orig_get_mvalues_val)( int );
obj_t (*bgl_orig_set_mvalues_val)( int, obj_t );

struct befored *(*bgl_orig_get_befored_top)();
void (*bgl_orig_set_befored_top)( struct befored * );

/*---------------------------------------------------------------------*/
/*    static char *                                                    */
/*    bglthread_get_stack_bottom ...                                   */
/*---------------------------------------------------------------------*/
static char *
bglthread_get_stack_bottom() {
   bglthread_t t = pthread_current_thread();

   return (t ? t->stack_bottom : bgl_orig_get_stack_bottom());
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    bglthread_set_stack_bottom ...                                   */
/*---------------------------------------------------------------------*/
static void
bglthread_set_stack_bottom( char * val ) {
   bglthread_t t = pthread_current_thread();

   if( t )
      t->stack_bottom = val;
   else
      bgl_orig_set_stack_bottom( val );
}

/*---------------------------------------------------------------------*/
/*    static obj_t                                                     */
/*    bglthread_get_exitd_top ...                                      */
/*---------------------------------------------------------------------*/
static obj_t
bglthread_get_exitd_top() {
   bglthread_t t = pthread_current_thread();

   return (t ? t->exitd_top : bgl_orig_get_exitd_top());
}

/*---------------------------------------------------------------------*/
/*    static obj_t                                                     */
/*    bglthread_set_exitd_top ...                                      */
/*---------------------------------------------------------------------*/
static obj_t
bglthread_set_exitd_top( obj_t val ) {
   bglthread_t t = pthread_current_thread();

   if( t )
      return (t->exitd_top = val), BUNSPEC;
   else
      return bgl_orig_set_exitd_top( val );
}

/*---------------------------------------------------------------------*/
/*    static obj_t                                                     */
/*    bglthread_get_current_output_port ...                            */
/*---------------------------------------------------------------------*/
static obj_t
bglthread_get_current_output_port() {
   bglthread_t t = pthread_current_thread();

   return (t ? t->current_output_port : bgl_orig_get_current_output_port());
}

/*---------------------------------------------------------------------*/
/*    static obj_t                                                     */
/*    bglthread_set_current_output_port ...                            */
/*---------------------------------------------------------------------*/
static obj_t
bglthread_set_current_output_port( obj_t val ) {
   bglthread_t t = pthread_current_thread();

   if( t )
      return (t->current_output_port = val), BUNSPEC;
   else
      return bgl_orig_set_current_output_port( val );
}

/*---------------------------------------------------------------------*/
/*    static obj_t                                                     */
/*    bglthread_get_current_error_port ...                             */
/*---------------------------------------------------------------------*/
static obj_t
bglthread_get_current_error_port() {
   bglthread_t t = pthread_current_thread();

   return (t ? t->current_error_port : bgl_orig_get_current_error_port());
}

/*---------------------------------------------------------------------*/
/*    static obj_t                                                     */
/*    bglthread_set_current_error_port ...                             */
/*---------------------------------------------------------------------*/
static obj_t
bglthread_set_current_error_port( obj_t val ) {
   bglthread_t t = pthread_current_thread();

   if( t )
      return (t->current_error_port = val), BUNSPEC;
   else
      return bgl_orig_set_current_error_port( val );
}

/*---------------------------------------------------------------------*/
/*    static obj_t                                                     */
/*    bglthread_get_current_input_port ...                             */
/*---------------------------------------------------------------------*/
static obj_t
bglthread_get_current_input_port() {
   bglthread_t t = pthread_current_thread();

   return (t ? t->current_input_port : bgl_orig_get_current_input_port());
}

/*---------------------------------------------------------------------*/
/*    static obj_t                                                     */
/*    bglthread_set_current_input_port ...                             */
/*---------------------------------------------------------------------*/
static obj_t
bglthread_set_current_input_port( obj_t val ) {
   bglthread_t t = pthread_current_thread();

   if( t )
      return (t->current_input_port = val), BUNSPEC;
   else
      return bgl_orig_set_current_input_port( val );
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    bglthread_get_mvalues_number ...                                 */
/*---------------------------------------------------------------------*/
static int
bglthread_get_mvalues_number() {
   bglthread_t t = pthread_current_thread();

   return (t ? t->mvalues_number : bgl_orig_get_mvalues_number());
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    bglthread_set_mvalues_number ...                                 */
/*---------------------------------------------------------------------*/
static int
bglthread_set_mvalues_number( int val ) {
   bglthread_t t = pthread_current_thread();

   if( t )
      return (t->mvalues_number = val), val;
   else
      return bgl_orig_set_mvalues_number( val );
}

/*---------------------------------------------------------------------*/
/*    static obj_t                                                     */
/*    bglthread_get_mvalues_val ...                                    */
/*---------------------------------------------------------------------*/
static obj_t
bglthread_get_mvalues_val( int n ) {
   bglthread_t t = pthread_current_thread();

   return (t ? t->mvalues_values[ n ] : bgl_orig_get_mvalues_val( n ));
}

/*---------------------------------------------------------------------*/
/*    static obj_t                                                     */
/*    bglthread_set_mvalues_val ...                                    */
/*---------------------------------------------------------------------*/
static obj_t
bglthread_set_mvalues_val( int n, obj_t val ) {
   bglthread_t t = pthread_current_thread();

   if( t )
      return (t->mvalues_values[ n ] = val), val;
   else
      return bgl_orig_set_mvalues_val( n, val );
}

/*---------------------------------------------------------------------*/
/*    static struct befored *                                          */
/*    bglthread_get_befored_top ...                                    */
/*---------------------------------------------------------------------*/
static struct befored *
bglthread_get_befored_top() {
   bglthread_t t = pthread_current_thread();

   return (t ? t->befored_top : bgl_orig_get_befored_top());
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    bglthread_set_befored_top ...                                    */
/*---------------------------------------------------------------------*/
static void
bglthread_set_befored_top( struct befored *val ) {
   bglthread_t t = pthread_current_thread();

   if( t )
      t->befored_top = val;
   else
      bgl_orig_set_befored_top( val );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    bglthread_rts_init ...                                           */
/*---------------------------------------------------------------------*/
void
bglthread_rts_init() {
   bgl_orig_get_stack_bottom = bgl_get_stack_bottom;
   bgl_orig_set_stack_bottom = bgl_set_stack_bottom;
   
   bgl_orig_get_exitd_top = bgl_get_exitd_top;
   bgl_orig_set_exitd_top = bgl_set_exitd_top;
   
   bgl_orig_get_current_output_port = bgl_get_current_output_port;
   bgl_orig_set_current_output_port = bgl_set_current_output_port;
   
   bgl_orig_get_current_error_port = bgl_get_current_error_port;
   bgl_orig_set_current_error_port = bgl_set_current_error_port;
   
   bgl_orig_get_current_input_port = bgl_get_current_input_port;
   bgl_orig_set_current_input_port = bgl_set_current_input_port;
   
   bgl_orig_get_mvalues_number = bgl_get_mvalues_number;
   bgl_orig_set_mvalues_number = bgl_set_mvalues_number;
   
   bgl_orig_get_mvalues_val = bgl_get_mvalues_val;
   bgl_orig_set_mvalues_val = bgl_set_mvalues_val;
   
   bgl_orig_get_befored_top = bgl_get_befored_top;
   bgl_orig_set_befored_top = bgl_set_befored_top;
   
   bgl_get_stack_bottom = &bglthread_get_stack_bottom;
   bgl_set_stack_bottom = &bglthread_set_stack_bottom;
   
   bgl_get_exitd_top = &bglthread_get_exitd_top;
   bgl_set_exitd_top = &bglthread_set_exitd_top;
   
   bgl_get_befored_top = &bglthread_get_befored_top;
   bgl_set_befored_top = &bglthread_set_befored_top;

   bgl_get_current_output_port = &bglthread_get_current_output_port;
   bgl_set_current_output_port = &bglthread_set_current_output_port;
   
   bgl_get_current_error_port = &bglthread_get_current_error_port;
   bgl_set_current_error_port = &bglthread_set_current_error_port;
   
   bgl_get_current_input_port = &bglthread_get_current_input_port;
   bgl_set_current_input_port = &bglthread_set_current_input_port;
   
   bgl_get_mvalues_number = &bglthread_get_mvalues_number;
   bgl_set_mvalues_number = &bglthread_set_mvalues_number;
   
   bgl_get_mvalues_val = &bglthread_get_mvalues_val;
   bgl_set_mvalues_val = &bglthread_set_mvalues_val;
}
