;*=====================================================================*/
;*    serrano/prgm/project/bigloo/fthread/src/scheduler.scm            */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Wed Feb  6 14:39:43 2002                          */
;*    Last change :  Wed Jul 24 18:44:33 2002 (serrano)                */
;*    Copyright   :  2002 Manuel Serrano                               */
;*    -------------------------------------------------------------    */
;*    Fair threads scheduler                                           */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __ft_scheduler
   
   (import __ft_types
	   __ft_%types
	   __ft_%scheduler
	   __ft_%thread
	   __ft_thread
	   __ft_env
	   __ft_%env
	   __ft_signal
	   __ft_exception)
   
   (export (make-scheduler . env)
	   (current-scheduler)
	   (default-scheduler . scdl)
	   (with-scheduler ::scheduler ::procedure)
	   (scheduler-react! . ::obj)
	   (scheduler-start! . args)
	   (scheduler-terminate! . ::obj)
	   (scheduler-instant::int . ::obj)
	   (scheduler-broadcast ::scheduler event . val)
	   
	   ;; misc 
	   (broadcast event . val)))
	   
;*---------------------------------------------------------------------*/
;*    *current-scheduler* & *default-scheduler* ...                    */
;*---------------------------------------------------------------------*/
(define *current-scheduler* #f)
(define *default-scheduler* #f)

;*---------------------------------------------------------------------*/
;*    current-scheduler ...                                            */
;*---------------------------------------------------------------------*/
(define (current-scheduler) *current-scheduler*)

;*---------------------------------------------------------------------*/
;*    default-scheduler ...                                            */
;*---------------------------------------------------------------------*/
(define (default-scheduler . scdl)
   (cond
      ((null? scdl)
       *default-scheduler*)
      ((scheduler? (car scdl))
       (set! *default-scheduler* (car scdl))
       (car scdl))
      (else
       (error "default-scheduler" "Illegal scheduler" (car scdl)))))

;*---------------------------------------------------------------------*/
;*    with-scheduler ...                                               */
;*---------------------------------------------------------------------*/
(define (with-scheduler s thunk)
   (let ((old (default-scheduler)))
      (unwind-protect
	 (begin
	    (default-scheduler s)
	    (thunk))
	 (default-scheduler old))))

;*---------------------------------------------------------------------*/
;*    make-scheduler ...                                               */
;*---------------------------------------------------------------------*/
(define (make-scheduler . envs)
   (letrec ((s (instantiate::%scheduler
		  (%builtin (%bgl-thread-new (lambda () (schedule s))))
		  (name (gensym "scheduler"))
		  (env (append envs
			       (list (instantiate::%env))))
		  (%exc-handlers (list (lambda (x) x))))))
      ;; initialize the back pointer facility
      (thread-setup! s)
      ;; start the native thread (that will block instantly)
      (%native-thread-start (thread-%builtin s))
      ;; if there is no default scheduler, store that one
      (if (not (default-scheduler)) (default-scheduler s))
      ;; return the newly allocates scheduler
      s))

;*---------------------------------------------------------------------*/
;*    scheduler-react! ...                                             */
;*---------------------------------------------------------------------*/
(define (scheduler-react! . s)
   (let ((s (cond
	       ((null? s)
		(default-scheduler))
	       ((scheduler? (car s))
		(car s))
	       (else
		(error "scheduler-react!"
		       "Illegal scheduler"
		       (car s))))))
      (with-access::%scheduler s (tostart deadn startn toterminate threads-yield tobroadcast)
	 (let ((o (current-scheduler)))
	    
	    (set! *current-scheduler* s)
	    (if (> *thread-debug* 3) (print "Switch in scheduler " s))
	    (thread-switch-in s)
	    (if (> *thread-debug* 3) (print "Came back from scheduler " s))
	    (set! *current-scheduler* o)
	    (cond
	       ((pair? tostart)
		'ready)
	       ((=fx deadn startn)
		'done)
	       ((>=fx (+fx deadn (length toterminate)) startn)
		'ready)
	       ((pair? tobroadcast)
		'ready)
	       ((null? threads-yield)
		'await)
	       (else
		'ready))))))

;*---------------------------------------------------------------------*/
;*    scheduler-react-debug ...                                        */
;*---------------------------------------------------------------------*/
(define (scheduler-react-debug s res)
   ;; a simple debugging function used to display the result
   ;; of each reaction
   (print "scheduler-react!: " res)
   (if (and (> *thread-debug* 1)
	    (eq? res 'await)
	    (%env? (car (scheduler-env s))))
       (begin
	  (print "           await: ")
	  (for-each (lambda (sgn)
		       (print "               -- "
			      (%signal-id sgn) ": ")
		       (print "                  "
			      (%signal-threads sgn))
		       (newline))
		    (%env-waiting-signals
		     (car (scheduler-env s))))
	  (newline))))

;*---------------------------------------------------------------------*/
;*    scheduler-start! ...                                             */
;*---------------------------------------------------------------------*/
(define (scheduler-start! . args)
   (let ((iterp (and (pair? args) (number? (car args))))
	 (next (cond
		  ((null? args)
		   (lambda (i)
		      (+fx i 1)))
		  ((number? (car args))
		   (lambda (i)
		      (if (< i (car args))
			  (+fx i 1)
			  #f)))
		  ((procedure? (car args))
		   (lambda (i)
		      (let ((res ((car args) i)))
			 (if res (+fx i 1) #f))))
		  (else
		   (error "scheduler-start!"
			  "Illegal optional parameter"
			  args))))
	 (s (cond
	       ((or (null? args) (null? (cdr args)))
		(default-scheduler))
	       ((scheduler? (cadr args))
		(cadr args))
	       (else
		(error "scheduler-start!" "Illegal scheduler" (cadr args))))))
      (let loop ((i 1))
	 (if i
	     (let ((res (scheduler-react! s)))
		(if (> *thread-debug* 0) (scheduler-react-debug s res))
		(case res
		   ((await)
		    (if (not iterp)
			(%bgl-asynchronous-thunk-funcall
			 (lambda ()
			    (begin
			       (if (not (pair? (%scheduler-tobroadcast s)))
				   (begin
				      (%bgl-asynchronous-thunk-await! s)
				      #unspecified))))
			 s))
		    (loop (next i)))
		   ((ready)
		    (loop (next i)))
		   (else
		    res)))))))

;*---------------------------------------------------------------------*/
;*    scheduler-terminate! ...                                         */
;*---------------------------------------------------------------------*/
(define (scheduler-terminate! . s)
   (let ((s (cond
	       ((null? s)
		(default-scheduler))
	       ((scheduler? (car s))
		(car s))
	       (else
		(error "scheduler-react!"
		       "Illegal scheduler"
		       (car s))))))
      (with-access::%scheduler s (tosuspend toresume tostart threads)
	 (set! toresume '())
	 (set! tosuspend '())
	 (set! tostart '())
	 (for-each thread-terminate! threads))))

;*---------------------------------------------------------------------*/
;*    scheduler-broadcast ...                                          */
;*---------------------------------------------------------------------*/
(define (scheduler-broadcast s event . val)
   (if (> *thread-debug* 2)
       (print "scheduler-broadcast: " event " " val))
   (%bgl-asynchronous-thunk-funcall
    (lambda ()
       (if (> *thread-debug* 0)
	   (begin
	      (display "%%% BROADCASTING: " (current-error-port))
	      (display event (current-error-port))))
       (with-access::%scheduler s (tobroadcast)
	  (set! tobroadcast (cons (cons event
					(if (pair? val)
					    (car val)
					    #unspecified))
				  tobroadcast)))
       (if (> *thread-debug* 0) (fprint (current-error-port) "...done")))
    s))

;*---------------------------------------------------------------------*/
;*    scheduler-instant ...                                            */
;*---------------------------------------------------------------------*/
(define (scheduler-instant . s)
   (let ((s (cond
	       ((null? s)
		(default-scheduler))
	       ((scheduler? (car s))
		(car s))
	       (else
		(error "scheduler-react!"
		       "Illegal scheduler"
		       (car s))))))
      (with-access::%scheduler s (env)
	 (ftenv-instant (car env)))))

;*---------------------------------------------------------------------*/
;*    broadcast ...                                                    */
;*---------------------------------------------------------------------*/
(define (broadcast event . val)
   (let ((t (current-thread)))
      (if (thread? t)
	  (if (thread-attached? t)
	      (with-access::thread t (scheduler)
		 (with-access::%scheduler scheduler (env)
		    (signal-emit event
				 (if (pair? val)
				     (car val)
				     #unspecified)
				 env)))
	      (error "broadcast" "Unattached thread" t)))))
   
;*---------------------------------------------------------------------*/
;*    schedule ...                                                     */
;*---------------------------------------------------------------------*/
(define (schedule scdl::scheduler)
   (if (> *thread-debug* 3) (print "schedule: " scdl))
   (define (init-instant)
      (with-access::%scheduler scdl (threads-runnable threads-yield env eoi)
	 ;; prepare the environment
	 (for-each instant-env! env)
	 ;; prepare the scheduler
	 (set! eoi #f)
	 (set! threads-runnable threads-yield)
	 (set! threads-yield '())))
   (define (dump-threads threads i state)
      ;; dump thread (this is a debug function)
      (if (> *thread-debug* 0)
	  (with-output-to-port (current-error-port)
	     (lambda ()
		(let ((m (make-string 30 #\space)))
		   (print m "--- " i " " state ":")
		   (if (> *thread-debug* 1)
		       (for-each (lambda (t)
				    (print m "  " t 
					   " [" (thread-%state t) "]"))
				 threads)
		       (print m "thread runnable: "  (length threads)))
		   (print m "------------------------------------ "))))))
   (define (run-threads threads state)
      ;; run all the threads
      (let ((thds (sort-threads threads)))
	 (dump-threads thds (ftenv-instant (car (scheduler-env scdl))) state)
	 (for-each (lambda (t)
		      (%scheduler-current-thread-set! scdl t)
		      (thread-switch scdl t))
		   thds)))
   (define (start-threads)
      (with-access::%scheduler scdl (tostart threads-runnable)
	 ;; append the new threads at the end of the threads list
	 (set! threads-runnable (append! (reverse! tostart) threads-runnable))
	 (set! tostart '())))
   (define (terminate-threads)
      (with-access::%scheduler scdl (toterminate threads-runnable)
	 ;; we have to remove multiple occurrences of thread to be
	 ;; terminated other we will have a lock because we will run
	 ;; a terminated thread
	 (let ((thds (filter (lambda (t)
				(with-access::thread t (%state)
				   (if (not (memq %state '(dying dead)))
				       (begin
					  (set! %state 'dying)
					  #t)
				       #f)))
			     toterminate)))
	    (set! toterminate '())
	    ;; mark that the thread is dying and remove it from the
	    ;; runnable and await list
	    (for-each (lambda (t)
			 (set! threads-runnable (remq! t threads-runnable)))
		      thds)
	    (run-threads thds 'dying))))
   (define (suspend-threads)
      (with-access::%scheduler scdl (tosuspend threads-suspend)
	 (set! threads-suspend (append! tosuspend threads-suspend))
	 (set! tosuspend '())))
   (define (resume-threads)
      (with-access::%scheduler scdl (toresume threads-suspend threads-runnable)
	 ;; append the new threads at the end of the threads list
	 (for-each (lambda (th)
		      (set! threads-suspend (remq! th threads-suspend))
		      (set! threads-runnable (cons th threads-runnable)))
		   toresume)
	 (set! toresume '())))
   (define (broadcast-events)
      (%bgl-asynchronous-thunk-funcall
       (lambda ()
	  (with-access::%scheduler scdl (tobroadcast env)
	     ;; generate all the signals
	     (for-each (lambda (x)
			  (signal-emit (car x) (cdr x) env))
		       tobroadcast)
	     (set! tobroadcast '())))
       scdl))
   (define (execute-threads)
      (with-access::%scheduler scdl (current-thread
				     env eoi
				     threads-runnable threads-eoi)
	 (let loop ()
	    ;; the runnable threads
	    (let ((ths threads-runnable))
	       (set! threads-runnable '())
	       (run-threads ths 'runnable))
	    ;; we check if there still are runnable threads
	    (if (pair? threads-runnable)
		(loop)
		(begin
		   (set! current-thread #f)
		   (set! eoi #t)
		   (let ((ths threads-eoi))
		      (set! threads-eoi '())
		      (run-threads ths 'eoi)))))))
   ;; prepare the threads and the instant numbering
   (if (> *thread-debug* 0)
       (print "\n" (make-string 29 #\=) " instant: "
	      (ftenv-instant (car (scheduler-env scdl)))
	      " ========================="))
   (init-instant)
   ;; add all the new threads
   (start-threads)
   ;; resume suspended threds
   (resume-threads)
   ;; suspend threads
   (suspend-threads)
   ;; prepare the threads that are about to die
   (terminate-threads)
   ;; broadcast the events
   (broadcast-events)
   ;; run until the end of the install
   (execute-threads)
   ;; leave the Jvm thread implementing this scheduler
   (thread-switch-out scdl)
   ;; when the scheduler will get the CPU, it will have to loop,
   ;; that is, to restart its attached threads.
   (schedule scdl))

